//JsBridge原理
//Native只开放request和response两个接口给Web调用
//两端均采用key-value的方式存储通信用的接口。即通信通道唯一（只有request和response），通信接口自定义（通过map存储）
//Web调用Native的方法，Native在接收正确处理后，根据Web传递过来的Callbackid，触发Web的对应回调。
//Native调用Web也是同理，Web在接收到方法调用并正确处理后，通过response，以及callbackey，触发Native的结果接收。

(function() {
    if (window.WebViewJavascriptBridge) {
        return;
    }
//  1--
//  javascript 自带的 typeof 运算符不能很好的显示所有的类型;
//  比如 typeof null; typeof [1,2,3];
//  下面函数的基本原理是调用 Object.prototype 原型上的 toString 方法;
    function typeOf_(obj) {
        const toString = Object.prototype.toString;
        const map = {
            '[object Boolean]': 'boolean',
            '[object Number]': 'number',
            '[object String]': 'string',
            '[object Function]': 'function',
            '[object Array]': 'array',
            '[object Date]': 'date',
            '[object RegExp]': 'regExp',
            '[object Undefined]': 'undefined',
            '[object Null]': 'null',
            '[object Object]': 'object',
            '[object Symbol]': 'symbol',
            '[object Error]': 'error',
        };
        return map[toString.call(obj)];
    }
    function isBoolean(args) {
      return typeOf_(args) === 'boolean';
    }
    function isNumber(args) {
      return typeOf_(args) === 'number';
    }
    function isString(args) {
      return typeOf_(args) === 'string';
    }
    function isArray(args) {
      return typeOf_(args) === 'array';
    }
    function isDate(args) {
      return typeOf_(args) === 'date';
    }
    function isRegExp(args) {
      return typeOf_(args) === 'regexp';
    }
    function isUndefined(args) {
      return typeOf_(args) === 'undefined';
    }
    function isNull(args) {
      return typeOf_(args) === 'null';
    }
    function isObject(args) {
      return typeOf_(args) === 'object';
    }
    function isSymbol(args) {
      return typeOf_(args) === 'symbol';
    }
    function isError(args) {
      return typeOf_(args) === 'error';
    }
    //Web调用Native时，Web接收Native回复结果时的方法
    var responseCallbacks = {};
    var uniqueId = 1;
    //web注册供Native调用的方法
    var messageHandlers = {};

    function registerHandler(handlerName, handler) {
        messageHandlers[handlerName] = handler;
    }

    function callHandler(handlerName, data, responseCallback) {
        _doSend({
            handlerName: handlerName,
            data: data
        }, responseCallback);
    }

    function fromNativeCallback(cbId, result) {
        if (cbId) {
            var responseCallback = responseCallbacks[cbId];
            if (responseCallback) {
                var decodeRet = decodeURIComponent(result);
                responseCallback(decodeRet);
                //TODO delete是否有问题？触发一次就移除？
                delete responseCallbacks[cbId];
            }
        }
    }

    function _handleMessageFromNative(handlerName, data, callbackKey) {
        var handler = messageHandlers[handlerName];
        if (handler) {
            var decodeParam = '';
            if (data) {
                decodeParam = decodeURIComponent(data);
            }
            handler(decodeParam, function(ret) {
                if (window.dwbridge) {
                    var retStr;
                    //最好通信结果全部走Json
                    if (isNumber(ret)) {
                        retStr = ret.toString();
                    } else if (isBoolean(ret)) {
                        retStr = ret.toString();
                    } else if (isString(ret)) {
                        retStr = ret;
                    } else {
                        retStr = JSON.stringify(ret);
                    }
                    window.dwbridge.response(retStr, callbackKey);
                }
            });
        }
    }

    //Web调用Native方法
    function _doSend(message, responseCallback) {
        if (responseCallback) {
            var callbackId = 'cb_' + (uniqueId++) + '_' + new Date().getTime();
            responseCallbacks[callbackId] = responseCallback;
            message.callbackId = callbackId;
        }
        if (window.dwbridge) {
            window.dwbridge.request(JSON.stringify(message));
        }
    }
    var WebViewJavascriptBridge = window.WebViewJavascriptBridge = {
        //Web调用Native的方法
        callHandler: callHandler,
        //Web注册开放方法，供Native调用
        registerHandler: registerHandler,
        //Web接收Native回调结果的方法
        fromNativeCallback: fromNativeCallback,
        //Web接收Native调用的地方
        _handleMessageFromNative: _handleMessageFromNative,
    };

    var doc = document;
    var readyEvent = doc.createEvent('Events');

    readyEvent.initEvent('WebViewJavascriptBridgeReady');
    readyEvent.bridge = WebViewJavascriptBridge;
    doc.dispatchEvent(readyEvent);
})();